/*
 * ping.c - Meine eigene Version von ping
 *
 * Copyright (c) 2000, 2002, 2010, Felix Opatz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * Meine Implementierung von ping(8) untersttzt die wichtigsten Optionen, die
 * von diversen anderen pings bekannt sind. Es wird ausschlielich IPv4
 * untersttzt. Zielplattform ist Linux 2.6, aber es sollte auch auf anderen
 * UNIX-artigen Plattformen lauffhig sein.

 * ping [-n] [-c count] [-i interval] [-s packetsize] [-W timeout] destination
 *
 * -h     Hilfe ausgeben und beenden.
 * -n     Nur numerische Ausgaben, keine Auflsung von Hostnamen
 * -c     Bestimmte Anzahl von Paketen senden bzw. empfangen, danach beenden.
 *        Default ist ewiges Senden bis der Benutzer unterbricht.
 * -i     Intervall zwischen den einzelnen Sendevorgngen, in Sekunden.
 *        Default ist eine Sekunde.
 * -s     Gre der Nutzdaten jedes Pakets, in Bytes.
 *        Default ist 56 Byte (zuzglich ICMP-Header 64 Byte).
 * -W     Timeout nach dem ein Paket als verloren gilt, in Sekunden.
 *        Default ist 30 Sekunden.
 *
 * Ziel kann eine IP-Adresse in Punktnotation oder ein Hostname sein.
 *
 *
 * Hinweise zur berarbeiteten Implementierung
 * ===========================================
 *
 * In der ursprnglichen Implementierung wurde das Senden der Ping-Anfragen
 * aus einem SIGALRM-Handler durchgefhrt. Das Hauptprogramm hat auf den
 * Empfang der Antworten gewartet. Dies hatte einige Nachteile (viel Code im
 * Signal-Handler, kein berblick zu den ausstehenden Antworten, dadurch keine
 * Timeout-Zeiten grer als die Ping-Frequenz mglich, etc.), die mit der
 * neuen Implementierung beseitigt wurden. Die neue Implementierung nutzt eine
 * Hauptschleife, in der das Senden und Empfangen abgewickelt wird. Nach dem
 * Senden einer Anfrage wird ein Eintrag in eine Liste eingefgt. Dieser
 * beinhaltet den Zeitpunkt des Absendens und die Sequenznummer. Wenn eine
 * Ping-Antwort eintrifft, wird der entsprechende Eintrag aus der Liste
 * entfernt und die Laufzeit ausgewertet. In der Liste verbliebene Eintrge
 * zeigen verlorene Pakete auf, und werden nach Ablauf des Timeouts entfernt.
 *
 * Um auf CTRL+C zu reagieren wird ein Signal-Handler auf SIGINT registriert.
 * Es wird bewusst nicht SA_RESTART gesetzt, um aus Systemaufrufen sofort
 * zurckkehren zu knnen. Die naheliegende Lsung, im Signal-Handler ein
 * Flag zu setzen, das im Kopf der Sendeschleife geprft wird, fhrt zu einer
 * Wettlaufsituation: wenn das Signal zwischen Schleifenkopf und select()
 * auftritt, wird potenziell ewig in select() blockiert. Dieses Problem wird
 * durch den Einsatz des 'self-pipe trick' umgangen. Siehe dazu auch
 * http://cr.yp.to/docs/selfpipe.html. Hierbei wird eine zustzliche Pipe
 * angelegt und mit in das Deskriptor-Set fr select() gesteckt. Wenn jetzt
 * aus dem Signal-Handler in die Pipe geschrieben wird, kehrt der select()-
 * Aufruf garantiert zurck. Die Lesbarkeit der Pipe wird als Hinweis darauf
 * gewertet, da der Signal-Handler durchlaufen wurde. Damit ist kein weiteres
 * Flag notwendig.
 *
 *
 * Datum       nderung
 * ---------------------------------------------------------------------------
 * 2010-05-30  Kleine Anpassungen an NetBSD.
 * 2010-05-15  Komplette berarbeitung mit besserer Paketverwaltung.
 * 2002-05-11  Berechnung der Checksumme korrigiert.
 * 2000-??-??  Ursprngliche Implementierung.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <unistd.h>
#include <fcntl.h>


/* Ein IPv4-Datagramm kann maximal 65535 Byte umfassen. Abzglich des IPv4-
 * Headers (20 Byte) und des ICMP-Headers (8 Byte) ergibt sich eine maximale
 * Nutzlast von 65535 - 28 = 65507 Byte.
 */
#define MAX_DGRAM_SIZE    65535
#define IPHDR_LEN         20
#define ICMPHDR_LEN       8
#define MAX_PAYLOAD_SIZE  (MAX_DGRAM_SIZE - IPHDR_LEN - ICMPHDR_LEN)

// Default-Werte fr Kommandozeilen-Parameter
#define COUNT_INFINITE    -1
#define DEF_COUNT         COUNT_INFINITE
#define DEF_INTERVAL      1000000ULL
#define DEF_PAYLOAD_SIZE  56
#define DEF_TIMEOUT       30000000ULL

#define MIN_VAL(a, b)     ((a) < (b) ? (a) : (b))
#define MAX_VAL(a, b)     ((a) > (b) ? (a) : (b))

/* Zeiten werden als 64-Bit-Timestamp in Mikrosekunden seit 1.1.1970
 * dargestellt. Dies bietet die Auflsung der struct timeval und den Vorteil
 * der leichten arithmetischen Verarbeitung einer Ganzzahl.
 */
typedef unsigned long long timestamp_t;

// Rckgabewerte fr wait_for_reply()
enum { WAIT_FAILED, WAIT_EXPIRED, WAIT_ABORTED, WAIT_SUCCEEDED };

// Konfiguration von Kommandozeile. Zeiten in Mikrosekunden.
struct config
{
	const char  *target;   // Ziel das angepingt wird
	int          help;     // Hilfe ausgeben und beenden
	int          numeric;  // Keine Namensauflsung
	int          count;    // Anzahl zu sendender Pakete
	int          size;     // Gre der Nutzdaten
	timestamp_t  interval; // Intervall zwischen Sendungen
	timestamp_t  timeout;  // Timeout bis Paket verloren gilt
};

// ICMP Echo Request, kann Teil einer doppelt verketteten Liste sein
struct echo_request
{
	uint16_t             seq;       // Sequenz-Nummer
	timestamp_t          timestamp; // Timestamp
	unsigned long        len;       // Lnge der Nutzdaten
	struct in_addr       addr;      // IP-Adresse des Ziels
	struct echo_request *prev;      // Vorheriges Listenelement
	struct echo_request *next;      // Nchstes Listenelement
};

/* Verwaltungs-Struktur fr doppelt verkettete Liste. Die tatschliche
 * Lnge ist unwichtig, es wird nur geprft ob die Liste leer ist.
 */
struct request_list
{
	struct echo_request *head; // Erstes Listenelement
	struct echo_request *tail; // Letztes Listenelement
};

// ICMP Echo Reply
struct echo_reply
{
	uint16_t        seq;       // Sequenz-Nummer
	timestamp_t     timestamp; // Timestamp
	unsigned long   len;       // Lnge des IP-Datagramms
	struct in_addr  addr;      // IP-Adresse des Absenders
	int             ttl;       // TTL-Wert bei Empfang
	timestamp_t     rtt;       // Errechnete Round-Trip-Time
};

// Verwaltungs-Struktur fr einen Ping-Vorgang
struct ping_session
{
	// Verwaltungsdaten
	int             sockfd;     // Socket-Deskriptor
	char           *name;       // Name des Ziels
	struct in_addr  addr;       // IP-Adresse des Ziels
	uint16_t        next_seq;   // Nchste Sequenznummer

	// Statistische Daten
	timestamp_t     start_time; // Startzeit des Pingens
	timestamp_t     end_time;   // Endzeit des Pingens
	unsigned long   tx_count;   // Anzahl gesendeter Pakete
	unsigned long   rx_count;   // Anzahl empfangener Pakete
	timestamp_t     rtt_sum;    // Summe aller RTTs
	timestamp_t     rtt_min;    // Minimale RTT
	timestamp_t     rtt_max;    // Maximale RTT
};


// Die Prozess-ID wird zur Identifikation der eigenen Pings genutzt
static pid_t process_id;
// Pipe fr den eingangs erwhnten 'self-pipe trick'
static int quit_pipe[2];


// Signal-Handler fr SIGINT
static void sigint_handler(int signum)
{
	write(quit_pipe[1], "Q", 1);
}

/* Berechnung der Checksumme fr den ICMP-Header gem RFC792:
 *
 *    The 16 bit one's complement of the one's complement sum of all 16
 *    bit words in the header.  For computing the checksum, the checksum
 *    field should be zero.
 *
 * Auf einer Maschine mit 2er-Komplement mu bei der Addition der bertrag auf
 * das untere 16-Bit-Wort addiert werden. Eine gute Erklrung des Algorithmus
 * befindet sich in RFC1071 Computing the Internet Checksum.
 */
static uint16_t in_cksum(const void *addr, size_t len)
{
	uint32_t sum = 0;
	uint16_t *curr_word = (uint16_t*) addr;

	while (len > 1)
	{
		sum += *curr_word++;
		len -= 2;
	}

	if (len > 0)
		sum += *(uint8_t*)curr_word;

	while (sum >> 16)
		sum = (sum & 0xFFFF) + (sum >> 16);

	return ~sum;
}

// Ermittelt aktuelle Uhrzeit in Mikrosekunden seit 1.1.1970, 00:00:00 UTC
static timestamp_t get_timestamp(void)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return (timestamp_t)tv.tv_sec * 1000000 + tv.tv_usec;
}

// Umwandlung eines Mikrosekunden-Timestamps in eine Struktur timeval
static void timestamp_to_timeval(timestamp_t in, struct timeval *out)
{
	out->tv_sec  = in / 1000000;
	out->tv_usec = in % 1000000;
}

// Hngt eine Kopie des bergebenen Requests an das Ende der Liste an
static int append_req(struct request_list *list, struct echo_request *req)
{
	struct echo_request *dup = malloc(sizeof(*dup));

	if (!dup)
		return -1;

	*dup = *req;
	dup->prev = NULL;
	dup->next = NULL;

	if (list->tail)
	{
		// An vorhandene Liste anhngen
		list->tail->next = dup;
		dup->prev = list->tail;
		list->tail = dup;
	}
	else
	{
		// Liste ist noch leer
		list->head = dup;
		list->tail = dup;
	}

	return 0;
}

// Entfernt das angegebene Element aus der Liste und gibt den Speicher frei
static struct echo_request *remove_req(struct request_list *list,
	struct echo_request *req)
{
	struct echo_request *prev = req->prev;
	struct echo_request *next = req->next;

	if (prev)
		prev->next = next;
	else
		list->head = next;

	if (next)
		next->prev = prev;
	else
		list->tail = prev;

	free(req);
	return next;
}

/* Sendet ein ICMP Echo Request auf dem angegebenen Socket.
 * Der Timestamp-Wert der Request-Struktur wird direkt vor dem Senden mit
 * der aktuellen Uhrzeit berschrieben. Die Nutzdaten bestehen aus 0-Bytes.
 * Rckgabewert ist die Zahl der gesendeten Bytes, -1 im Fehlerfall.
 */
static int send_request(int sockfd, struct echo_request *req)
{
	unsigned char packet[MAX_DGRAM_SIZE];
	size_t total_len;
	struct icmp *icmp;
	struct sockaddr_in addr;
	int bytes;

	memset(packet, 0x00, sizeof(packet));
	icmp = (struct icmp*) packet;
	icmp->icmp_type = ICMP_ECHO;
	icmp->icmp_id = htons(process_id);
	icmp->icmp_seq = htons(req->seq);
	total_len = ICMPHDR_LEN + req->len;
	icmp->icmp_cksum = in_cksum(packet, total_len);

	addr.sin_addr = req->addr;
	addr.sin_family = AF_INET;
	addr.sin_port = 0;

	req->timestamp = get_timestamp();
	bytes = sendto(sockfd, packet, total_len, 0,
		(struct sockaddr*) &addr, sizeof(addr));

	return bytes;
}

/* Empfngt ein ICMP Echo Reply Datagramm auf dem angegebenen Socket.
 * Rckgabewert ist die Zahl der empfangenen Bytes, im Fehlerfall -1.
 * Fremde Datagramme werden verworfen, Rckgabewert ist dann 0 Bytes.
 */
static int recv_reply(int sockfd, struct echo_reply *reply)
{
	unsigned char packet[MAX_DGRAM_SIZE];
	struct ip *ip;
	struct icmp *icmp;
	int bytes;
	struct sockaddr_in from;
	socklen_t from_len;

	from_len = sizeof(from);
	bytes = recvfrom(sockfd, packet, sizeof(packet), 0,
		(struct sockaddr*) &from, &from_len);

	if (bytes == -1)
		return -1;
	
	ip = (struct ip*) packet;
	icmp = (struct icmp*) (packet + (ip->ip_hl * 4));

	if (icmp->icmp_type != ICMP_ECHOREPLY)
		return 0;

	if (ntohs(icmp->icmp_id) != process_id)
		return 0;

	reply->seq = ntohs(icmp->icmp_seq);
	reply->timestamp = get_timestamp();
	reply->len = bytes - (ip->ip_hl * 4);
	reply->ttl = ip->ip_ttl;
	reply->addr = from.sin_addr;

	return bytes;
}

/* Wartet auf das Eintreffen eines ICMP Echo Reply Datagramms.
 * Wenn der angegebene Zeitpunkt erreicht wurde, kehrt die Funktion in jedem
 * Fall zurck. Wenn das Warten erfolgreich war, enthlt der Parameter reply
 * die empfangene Antwort.
 *
 * Rckgabewerte:
 * WAIT_FAILED     Warten fehlgeschlagen
 * WAIT_EXPIRED    Wartezeit ist abgelaufen
 * WAIT_ABORTED    Wartevorgang wurde abgebrochen
 * WAIT_SUCCEEDED  Warten erfolgreich, Antwort empfangen
 */
static int wait_for_reply(int sockfd, timestamp_t exp_time,
	struct echo_reply *reply)
{
	fd_set fds;
	timestamp_t curr_time;
	int maxfd = MAX_VAL(sockfd, quit_pipe[0]);
	int result;
	struct timeval tv;

	curr_time = get_timestamp();
	while (curr_time < exp_time)
	{
		FD_ZERO(&fds);
		FD_SET(sockfd, &fds);
		FD_SET(quit_pipe[0], &fds);
		timestamp_to_timeval(exp_time - curr_time, &tv);
		result = select(maxfd + 1, &fds, NULL, NULL, &tv);

		if (result == -1 && errno != EINTR)
			return WAIT_FAILED;
		if (result == 0)
			return WAIT_EXPIRED;
		if (FD_ISSET(quit_pipe[0], &fds))
			return WAIT_ABORTED;
		if (FD_ISSET(sockfd, &fds))
		{
			result = recv_reply(sockfd, reply);
			if (result == -1)
				return WAIT_FAILED;
			if (result > 0)
				return WAIT_SUCCEEDED;
		}

		curr_time = get_timestamp();
	}

	return WAIT_EXPIRED;
}

// Gibt eine Empfangsbesttigung auf die Konsole aus
static void print_reply(const struct echo_reply *reply, int numeric)
{
	unsigned long len = reply->len;
	const char *addr = inet_ntoa(reply->addr);
	const char *name;

	if (!numeric)
	{
		struct hostent *host;
		host = gethostbyaddr((void*)&reply->addr,
			sizeof(reply->addr), AF_INET);
		name = host ? host->h_name : addr;
	}

	if (numeric)
		printf("%lu bytes from %s: ", len, addr);
	else
		printf("%lu bytes from %s (%s): ", len, name, addr);

	printf("icmp_seq=%i ttl=%i time=%.1f ms\n",
		reply->seq, reply->ttl, reply->rtt / 1000.0);
}

// Initialisiert eine Ping-Session
static int ping_init(struct ping_session *ps, const struct config *cfg)
{
	struct in_addr addr;
	struct hostent *host;

	memset(ps, 0, sizeof(*ps));
	ps->next_seq = 1;
	ps->rtt_min = -1ULL;

	/* Das angegebene Ziel kann ein Hostname oder eine IP-Adresse sein.
	 * Wenn es eine IP-Adresse ist, wird die Adresse gemerkt und der
	 * Hostname auf die Adresse in dotted-notation gesetzt.
	 * Wenn es ein Hostname ist, wird dieser aufgelst. Als Adresse wird
	 * die erste gelieferte Adresse (h_addr) genutzt, der Hostname wird
	 * auf den offiziellen Namen des Ziels (h_name) gesetzt.
	 */
	switch (inet_pton(AF_INET, cfg->target, &addr))
	{
	case -1:
		// Interner Fehler aufgetreten
		perror("Unable to translate target address.");
		return -1;
	case 0:
		// Ist ein Hostname
		host = gethostbyname(cfg->target);
		if (!host)
		{
			perror("Unable to resolve target hostname.");
			return -1;
		}

		ps->addr = *(struct in_addr*)host->h_addr;
		ps->name = strdup(host->h_name);
		break;
	case 1:
		// Ist eine IP-Adresse
		ps->addr = addr;
		ps->name = strdup(cfg->target);
		break;
	}

	return 0;
}

// Setzt die Startzeit des Ping-Vorgangs und gibt eine Zusammenfassung aus,
// welcher Host angepingt wird
static void ping_begin(struct ping_session *ps, const struct config *cfg)
{
	int payload_size = cfg->size;
	int datagram_size = payload_size +
		ICMPHDR_LEN + IPHDR_LEN;

	printf("PING %s (%s) %i(%i) bytes of data.\n",
		ps->name, inet_ntoa(ps->addr),
		payload_size, datagram_size);

	ps->start_time = get_timestamp();
}

// Hauptschleife fr das Senden und Empfangen der Pings
static int ping_loop(struct ping_session *ps, const struct config *cfg)
{
	int wait_result;
	struct echo_request request;
	struct echo_reply reply;
	struct request_list pending_reqs = {NULL};
	struct echo_request *req;
	timestamp_t curr_time;
	timestamp_t curr_exp;
	timestamp_t next_send = 0;
	timestamp_t latest_exp;
	timestamp_t next_event;
	int tx_left;

	request.len = cfg->size;
	request.addr = ps->addr;
	tx_left = cfg->count;

	/* tx_left wird fr unendlich langes Senden auf -1 gesetzt. Daher
	 * muss hier auf != 0 geprft werden. Nach dem Senden eines Pings
	 * darf tx_left nur dekrementiert werden, wenn es > 0 ist.
	 */
	while (tx_left || pending_reqs.head)
	{
		if (tx_left)
		{
			if (pending_reqs.head)
				next_event = MIN_VAL(latest_exp, next_send);
			else
				next_event = next_send;
		}
		else
			next_event = latest_exp;

		wait_result = wait_for_reply(ps->sockfd,
			next_event, &reply);

		if (wait_result == WAIT_FAILED)
			return -1;
		if (wait_result == WAIT_ABORTED)
			return 0;

		curr_time = get_timestamp();

		// Pings raussenden
		if (curr_time >= next_send && tx_left)
		{
			request.seq = ps->next_seq++;
			send_request(ps->sockfd, &request);
			if (append_req(&pending_reqs, &request) != 0)
				return -1;
			ps->tx_count++;
			if (tx_left > 0)
				tx_left--;
			next_send = curr_time + cfg->interval;
		}

		// Ausstehende Antworten durchsuchen
		req = pending_reqs.head;
		latest_exp = 0;
		while (req)
		{
			curr_exp = req->timestamp + cfg->timeout;
			if (curr_time >= curr_exp)
			{
				// Timeout abgelaufen
				req = remove_req(&pending_reqs, req);
				continue;
			}

			if ((wait_result == WAIT_SUCCEEDED) &&
				(req->seq == reply.seq))
			{
				// Antwort empfangen
				reply.rtt = reply.timestamp - req->timestamp;
				print_reply(&reply, cfg->numeric);
				ps->rx_count++;
				ps->rtt_sum += reply.rtt;
				ps->rtt_min = MIN_VAL(ps->rtt_min, reply.rtt);
				ps->rtt_max = MAX_VAL(ps->rtt_max, reply.rtt);
				req = remove_req(&pending_reqs, req);
				continue;
			}

			if (curr_exp > latest_exp)
				latest_exp = curr_exp;

			req = req->next;
		}
	}

	return 0;
}

// Setzt die Endzeit des Ping-Vorgangs und gibt die Statistik aus
static void ping_end(struct ping_session *ps)
{
	ps->end_time = get_timestamp();

	// Abbruch erfolgte noch vor dem ersten Senden
	if (ps->tx_count == 0)
		return;

	printf("\n--- %s ping statistics ---\n", ps->name);

	printf("%lu packets transmitted, %lu received, ",
		ps->tx_count, ps->rx_count);

	printf("%.0f%% packet loss, time %llums\n",
		(ps->tx_count - ps->rx_count) * 100.0 / ps->tx_count,
		(ps->end_time - ps->start_time) / 1000);
	
	if (ps->rx_count > 0)
		printf("rtt min/avg/max = %.3f/%.3f/%.3f ms\n",
			ps->rtt_min / 1000.0,
			ps->rtt_sum / (ps->rx_count * 1000.0),
			ps->rtt_max / 1000.0);
}

// Erzeugt eine Pipe und macht das Schreib-Ende nicht-blockierend
static int create_pipe(int pipefd[2])
{
	int flags;

	if (pipe(pipefd) == -1)
	{
		perror("Unable to create pipe");
		return -1;
	}

	flags = fcntl(pipefd[1], F_GETFL, 0);
	if (flags == -1)
	{
		perror("Unable to get descriptor flags");
		return -1;
	}

	flags |= O_NONBLOCK;
	if (fcntl(pipefd[1], F_SETFL, flags) == -1)
	{
		perror("Unable to set descriptor flags");
		return -1;
	}

	return 0;
}

// Installiert einen Signal-Handler fr das angegebene Signal. Es werden
// keine Flags gesetzt, insbesondere nicht SA_RESTART.
static int install_signal_handler(int signum, void (*handler)(int))
{
	sigset_t sigset;
	struct sigaction sa;

	sigemptyset(&sigset);
	sa.sa_mask = sigset;
	sa.sa_flags = 0;
	sa.sa_handler = handler;

	if (sigaction(signum, &sa, NULL) != 0)
	{
		perror("Unable to register signal handler");
		return -1;
	}

	return 0;
}

// Ermittelt die Kommandozeilen-Konfiguration
static int get_opts(int argc, char **argv, struct config *cfg)
{
	int failed = 0;
	int c;
	const char *getopt_str = "c:i:s:W:nh";

	while (!failed && (c = getopt(argc, argv, getopt_str)) != -1)
	{
		switch (c)
		{
		case 'c':
			cfg->count = atoi(optarg);
			if (cfg->count <= 0)
			{
				fprintf(stderr, "Invalid packet count.\n");
				failed = 1;
			}
			break;
		case 'i':
			cfg->interval = atoi(optarg) * 1000000ULL;
			if (cfg->interval <= 0)
			{
				fprintf(stderr, "Invalid interval.\n");
				failed = 1;
			}
			break;
		case 's':
			cfg->size = atoi(optarg);
			if (cfg->size <= 0)
			{
				fprintf(stderr, "Invalid packetsize.\n");
				failed = 1;
			}
			break;
		case 'W':
			cfg->timeout = atoi(optarg) * 1000000ULL;
			if (cfg->timeout <= 0)
			{
				fprintf(stderr, "Invalid timeout.\n");
				failed = 1;
			}
			break;
		case 'n':
			cfg->numeric = 1;
			break;
		case 'h':
			cfg->help = 1;
			break;
		case '?':
			failed = 1;
			break;
		default:
			fprintf(stderr, "Unknown error in getopt().\n");
			abort();
		}
	}

	// Der erste nachfolgende Parameter ist das Ziel
	if (optind < argc)
		cfg->target = argv[optind++];

	// Weitere Parameter sind Bedienungsfehler
	if (optind < argc)
		cfg->help = 1;

	return failed;
}

/* Einstiegspunkt. Hier wird davon Gebrauch gemacht, dass Sockets, Pipes und
 * Speicher beim Beenden implizit vom Betriebssystem freigegeben werden. Falls
 * das Programm in anderer Umgebung eingesetzt werden soll, mssten diese
 * Aufrumarbeiten selbst erledigt werden.
 */
int main(int argc, char *argv[])
{
	struct config cfg =
	{
		.count    = DEF_COUNT,
		.size     = DEF_PAYLOAD_SIZE,
		.interval = DEF_INTERVAL,
		.timeout  = DEF_TIMEOUT,
	};

	struct ping_session ps;
	int sockfd;

	// Wenn das Programm mit SUID-Bit als root gestartet wird, werden
	// diese Rechte nach dem ffnen des Sockets abgelegt.
	sockfd = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
	if (sockfd == -1)
	{
		perror("Unable to create socket");
		return EXIT_FAILURE;
	}
	seteuid(getuid());

	// Kommandozeilen-Konfiguration auswerten
	if (get_opts(argc, argv, &cfg) != 0)
		return EXIT_FAILURE;

	if (cfg.help)
	{
		printf("Usage: ping [-n] [-c count] [-i interval] [-s packetsize]\n");
		printf("            [-W timeout] destination\n");
		return EXIT_SUCCESS;
	}

	if (!cfg.target)
	{
		fprintf(stderr, "Missing destination.\n");
		return EXIT_FAILURE;
	}

	if (cfg.size > MAX_PAYLOAD_SIZE)
	{
		fprintf(stderr, "Packet size %i is too large. "
			"Maximum is %i.\n", cfg.size, MAX_PAYLOAD_SIZE);
		return EXIT_FAILURE;
	}

	// Pipe fr 'self-pipe trick' anlegen
	if (create_pipe(quit_pipe) != 0)
		return EXIT_FAILURE;

	// Signal-Handler fr SIGINT (Ctrl+C) installieren
	if (install_signal_handler(SIGINT, sigint_handler) != 0)
		return EXIT_FAILURE;

	if (ping_init(&ps, &cfg) != 0)
		return EXIT_FAILURE;

	ps.sockfd = sockfd;
	process_id = getpid();

	ping_begin(&ps, &cfg);
	if (ping_loop(&ps, &cfg) != 0)
		return EXIT_FAILURE;
	ping_end(&ps);

	return EXIT_SUCCESS;
}

